import time

from onestep import step
from onestep.broker import RabbitMQBroker, MemoryBroker
from loguru import logger

step.set_debugging()

# 配置RabbitMQ队列及连接信息，设置优先级为10
rmq_broker = RabbitMQBroker(
    "onstep_test_priority",
    {
        "username": "admin",
        "password": "admin",
    },
    queue_params={"arguments": {"x-max-priority": 10}},
)

# 模拟一个内存队列
todo_broker = MemoryBroker()


# 发送三个不同优先级的消息，其中一个默认优先级
# rmq_broker.send("1, priority=default")
# rmq_broker.send({"hello": "world", "priority": 5}, properties={"priority": 5})
# rmq_broker.send("3, priority=1", properties={"priority": 1})


@step(from_broker=rmq_broker, to_broker=todo_broker, workers=3)
def build_todo_list(message):
    logger.debug("build_todo_list", message.body)
    # 返回的内容将发给RabbitMQ队列
    return message


if __name__ == "__main__":
    step.start()
    time.sleep(5)
    step.shutdown()
    # rmq_broker.shutdown()
    pass



from loguru import logger

from onestep import step
from onestep.broker import MemoryBroker
from onestep.retry import TimesRetry, RetryIfException, AllRetry

todo_broker = MemoryBroker()


# todo_broker = RabbitMQBroker(
#     "test",
#     {
#         "username": "admin",
#         "password": "admin",
#     }
# )

@step(to_broker=todo_broker)
def build_todo_list():
    # mock data
    yield from range(10)


def callback_on_failure(message):
    # 这里可以按需记录到日志或者发送到失败队列
    logger.warning(f"failure_callback: will send to failure queue: {message}")
    logger.debug(message.to_json(True))
    pass


@step(
    from_broker=todo_broker,
    retry=AllRetry(TimesRetry(), RetryIfException((ValueError,))),
    error_callback=callback_on_failure
)
def do_something(message):
    logger.info(f"todo: {message.body}")
    # return
    # if message.body.get("id") == 1:
    #     raise RetryViaLocal("Invalid id")
    if message.body == 1:
        raise ValueError("Invalid id")


if __name__ == "__main__":
    step.set_debugging()
    build_todo_list()
    step.start(block=True)



from datetime import datetime
from typing import Any, Dict, List, Optional
from uuid import UUID

from sqlmodel import JSON, TypeDecorator
from onestep import step
from onestep.broker import MemoryBroker
from onestep.broker.mysql import MysqlBroker, Model, Field
from onestep.retry import AdvancedRetry

todo_broker = MemoryBroker()

class JSONEncodedDict(TypeDecorator):
    """自定义 JSON 类型装饰器，处理复杂对象序列化"""

    impl = JSON

    def process_bind_param(self, value: Any, dialect: Any) -> Any:
        if value is not None:
            return self._convert_for_json(value)
        return value

    def process_result_value(self, value: Any, dialect: Any) -> Any:
        return value

    @staticmethod
    def _convert_for_json(obj: Any) -> Any:
        if isinstance(obj, (list, tuple)):
            return [JSONEncodedDict._convert_for_json(v) for v in obj]
        elif isinstance(obj, dict):
            return {k: JSONEncodedDict._convert_for_json(v) for k, v in obj.items()}
        elif isinstance(obj, UUID):
            return str(obj)
        elif isinstance(obj, datetime):
            return obj.isoformat()
        elif hasattr(obj, "dict"):  # 处理 Pydantic/SQLModel 模型
            return JSONEncodedDict._convert_for_json(obj.dict())
        return obj


class ItemModel(Model, table=True):
    __tablename__ = "item"  # pyright: ignore[reportAssignmentType]
    
    id: int = Field(primary_key=True)
    title: str
    content: str
    status: str
    ids: Optional[List[int]] = Field(
        default=None,
        sa_type=JSONEncodedDict,
        sa_column_kwargs={"comment": "ID列表"},
    )
    meta: Optional[Dict] = Field(
        default=None, sa_type=JSONEncodedDict, sa_column_kwargs={"comment": "原始数据"}
    )
    created_at: datetime = Field(default_factory=datetime.now)


mysql_broker = MysqlBroker(params={
    "uri": "mysql+pymysql://root:root@localhost:3306/onestep"
})
@step(to_broker=todo_broker)
def build_todo_list():
    # mock data
    yield from [
        {
            "id": 1,
            "title": "todo1",
            "content": "todo1 content",
            "status": "todo",
            "ids": [1, 2, 3],
            "meta": {"key": "value"},
            "created_at": "2023-01-01 00:00:00",
        },
        {
            "id": 2,
            "title": "todo2",
            "content": "todo2 content",
            "status": "todo",
            "ids": [4, 5, 6],
            "meta": {"key": "value"},
            "created_at": "2023-01-02 00:00:00",    
        },
        {
            "id": 3,
            "title": "todo3",
            "content": "todo3 content",
            "status": "todo",
            "ids": [7, 8, 9],
            "meta": {"key": "value"},
            "created_at": "2023-01-03 00:00:00",    
        },
    ]


@step(from_broker=todo_broker, workers=1,
      to_broker=mysql_broker,
      retry=AdvancedRetry()
      )
def do_something(todo):
    todo.body["status"] = "done"
    if isinstance(todo.body.get("created_at"), str):
        try:
            todo.body["created_at"] = datetime.fromisoformat(todo.body["created_at"])  # noqa: F401
        except ValueError:
            todo.body["created_at"] = datetime.strptime(todo.body["created_at"], "%Y-%m-%d %H:%M:%S")
    return ItemModel(**todo.body)


if __name__ == '__main__':
    import time
    step.set_debugging()

    build_todo_list()
    step.start(block=False)
    time.sleep(5)
    step.shutdown()



from onestep import step, RedisPubSubBroker

broker = RedisPubSubBroker(channel="test")


@step(from_broker=broker)
def job_rds(message):
    print(message)


if __name__ == '__main__':
    step.set_debugging()
    step.start(block=True)



import json
import threading
from queue import Queue
from typing import Optional, Dict, Any, Callable, List, TYPE_CHECKING

from onestep.broker import BaseBroker, BaseConsumer
from onestep.message import Message
try:
    from use_sqs import SQSStore
except ImportError:
    SQSStore = None


class _SQSMessage(Message):
    """SQS消息类"""

    @classmethod
    def from_broker(cls, broker_message: Any):
        """从SQS消息创建Message对象"""
        required_attrs = ("body", "delete", "message_id")
        if not all(hasattr(broker_message, attr) for attr in required_attrs):
            raise TypeError(
                f"Message object missing required SQS attributes: {required_attrs}"
            )

        try:
            # 如果body已经是dict，直接使用；如果是字符串，则解析JSON
            body = broker_message.body
            if isinstance(body, dict):
                message = body
            else:
                message = json.loads(body)
        except (json.JSONDecodeError, TypeError):
            message = {"body": broker_message.body}

        return cls(
            body=message.get("body"), extra=message.get("extra"), message=broker_message
        )


class SQSBroker(BaseBroker):
    """SQS消息队列Broker实现"""

    message_cls = _SQSMessage

    def __init__(
        self,
        queue_name: str,
        message_group_id: str,
        message_deduplication_id_func: Optional[Callable[[Any], str]] = None,
        params: Optional[Dict] = None,
        prefetch: Optional[int] = 1,
        auto_create: bool = True,
        queue_params: Optional[Dict] = None,
        *args,
        **kwargs,
    ):
        """
        初始化SQS Broker

        :param queue_name: 队列名称
        :param message_group_id: 消息组ID (FIFO队列必需)
        :param message_deduplication_id_func: 接收 msg_body(json_str) 作为参数，返回 str 类型的 message_deduplication_id
        :param params: SQS连接参数
        :param prefetch: 预取消息数量
        :param auto_create: 是否自动创建队列
        :param queue_params: 队列参数
        """
        super().__init__(*args, **kwargs)
        self.queue_name = queue_name
        self.queue: Queue = Queue()
        self.prefetch = prefetch
        self.message_group_id = message_group_id
        self.message_deduplication_id_func = message_deduplication_id_func
        self.threads: List[threading.Thread] = []
        self._shutdown = False
        self._consuming_started = False
        self._consume_lock = threading.Lock()

        # 创建SQSStore实例
        if SQSStore is None:
            raise ImportError("Please install the `use-sqs` module to use SQSBroker")
        self.store = SQSStore(**(params or {}))

        # 确保队列存在
        if auto_create:
            self.store.declare_queue(queue_name, attributes=queue_params)

    def _consume(self, *args, **kwargs):
        """消费消息的内部方法"""
        prefetch = kwargs.pop("prefetch", self.prefetch)

        def callback(message: Any) -> None:
            """处理接收到的消息"""
            # 直接将原始SQS消息放入队列，保留完整的消息引用
            if self.queue is not None:
                self.queue.put(message)

        self.store.start_consuming(
            self.queue_name, callback=callback, prefetch=prefetch, **kwargs
        )

    def consume(self, *args, **kwargs) -> "SQSConsumer":
        """启动消费者"""
        daemon = kwargs.pop("daemon", True)
        timeout = kwargs.pop("timeout", 1000)
        with self._consume_lock:
            if not self._consuming_started:
                thread_kwargs = kwargs.copy()
                thread = threading.Thread(target=self._consume, args=args, kwargs=thread_kwargs)
                thread.daemon = daemon
                thread.start()
                self.threads.append(thread)
                self._consuming_started = True
        return SQSConsumer(self, timeout=timeout)

    def publish(self, message: Any, **kwargs):
        """发布消息"""
        if self.message_deduplication_id_func:
            message_deduplication_id = self.message_deduplication_id_func(message)
            if (
                isinstance(message_deduplication_id, str)
                and message_deduplication_id.strip()
                and len(message_deduplication_id.strip()) <= 128
            ):
                kwargs["message_deduplication_id"] = message_deduplication_id

        self.store.send(
            self.queue_name,
            message=message,
            message_group_id=self.message_group_id,
            **kwargs,
        )

    def confirm(self, message: Message):
        """确认消息"""
        broker_msg = getattr(message, "message", None)
        if broker_msg is not None and hasattr(broker_msg, "delete"):
            broker_msg.delete()

    def reject(self, message: Message):
        """拒绝消息"""
        broker_msg = getattr(message, "message", None)
        if broker_msg is not None and hasattr(broker_msg, "delete"):
            broker_msg.delete()

    def requeue(self, message: Message, is_source: bool = False):
        """
        重新入队消息

        :param message: 消息对象
        :param is_source: 是否使用原始消息
        """
        broker_msg = getattr(message, "message", None)
        # 确认原始消息，重新入队处理后的消息，以实现将消息放在队列尾部并等待处理
        if is_source and broker_msg is not None and hasattr(broker_msg, "body"):
            if hasattr(broker_msg, "delete"):
                broker_msg.delete()
            self.store.send(
                self.queue_name,
                broker_msg.body,
                message_group_id=self.message_group_id,
            )
        else:
            if broker_msg is not None and hasattr(broker_msg, "delete"):
                broker_msg.delete()
            # 使用 to_json() 序列化完整的 Message 对象
            self.store.send(
                self.queue_name, message.to_json(), message_group_id=self.message_group_id
            )

    def shutdown(self):
        """关闭Broker"""
        self._shutdown = True
        self.store.shutdown()
        for thread in self.threads:
            thread.join()
        self.queue = Queue()
        self.threads.clear()
        self._consuming_started = False


class SQSConsumer(BaseConsumer): ...



import threading

from onestep import MemoryBroker, step, BaseMiddleware


class MyMiddleware(BaseMiddleware):

    def before_send(self, step, message, *args, **kwargs):
        print("called before send")

    def after_send(self, step, message, *args, **kwargs):
        print("called after send")

    def before_consume(self, step, message, *args, **kwargs):
        print("called before receive")

    def after_consume(self, step, message, *args, **kwargs):
        print("called after receive")


memory_broker = MemoryBroker(once=True)
memory_broker2 = MemoryBroker()


@step(from_broker=memory_broker,
      to_broker=memory_broker2,
      middlewares=[MyMiddleware()])
def once_job(message):
    print(message)
    return message


threading.Thread(target=step.start, kwargs={'block': True}).start()

memory_broker.publish("the first message")



import json

import pytest
from onestep.broker import MemoryBroker

message = "message1"


@pytest.fixture
def broker():
    return MemoryBroker()


@pytest.fixture
def consumer(broker):
    broker.send(message)
    return broker.consume()


def test_memory_broker_send(broker, consumer):
    # broker.queue.get() 拿到的是原始消息字符串，一般不需要直接使用
    reached_message = broker.queue.get()
    assert json.loads(reached_message).get("body") == message


def test_memory_broker_consume(broker, consumer):
    # broker.consume() 拿到的是 Message 对象
    consumer = broker.consume()
    assert next(consumer).body == message


def test_memory_broker_reject(broker, consumer):
    msg = next(consumer)
    broker.reject(msg)
    assert next(consumer) is None


def test_memory_broker_requeue_source(broker, consumer):
    msg = next(consumer)
    msg.body = "message2"
    broker.requeue(msg, is_source=True)  # 原始消息重新入队
    assert next(consumer).body == message


def test_memory_broker_requeue_not_source(broker, consumer):
    msg = next(consumer)
    msg.body = "message2"
    broker.requeue(msg, is_source=False)  # 修改后的消息重新入队
    assert next(consumer).body == "message2"



from typing import Optional, Dict, Any

from onestep.broker import BaseBroker

try:
    from use_sqs import SNSPublisher
except ImportError:
    SNSPublisher = None


class SNSBroker(BaseBroker):
    """SNS消息Broker实现"""

    def __init__(
        self,
        topic_arn: str,
        message_group_id: Optional[str] = None,
        params: Optional[Dict] = None,
        *args,
        **kwargs,
    ):
        """
        初始化SNS Broker

        :param topic_arn: SNS主题ARN
        :param message_group_id: 消息组ID (FIFO主题必需)
        :param params: SNS连接参数
        """
        if SNSPublisher is None:
            raise ImportError("Please install the `use-sqs` module to use SNSBroker")

        super().__init__(*args, **kwargs)
        self.topic_arn = topic_arn
        self.message_group_id = message_group_id

        # 创建SNSPublisher实例
        self.publisher = SNSPublisher(**(params or {}))

    def publish(self, message: Any, **kwargs):
        """发布消息"""
        # 如果初始化时设置了message_group_id，且调用时未指定，则使用初始化的值
        if self.message_group_id and "message_group_id" not in kwargs:
            kwargs["message_group_id"] = self.message_group_id

        self.publisher.publish(
            self.topic_arn,
            message=message,
            **kwargs,
        )

    def consume(self, *args, **kwargs):
        """消费消息 - SNS不支持直接消费"""
        raise NotImplementedError("SNS Broker does not support consumption")



import time

from onestep import step
from onestep.broker import RedisStreamBroker, MemoryBroker

step.set_debugging()

# 配置队列及连接信息
rds_broker = RedisStreamBroker(stream="onestep_demo", params={})

# 模拟一个内存队列
todo_broker = MemoryBroker()


rds_broker.send("1hello world")
rds_broker.send({"2hello": "world"})

pass
# todo_broker.send("1")


@step(from_broker=rds_broker, to_broker=todo_broker, workers=1)
def build_todo_list(message):
    print("build_todo_list", message.body)
    return message


if __name__ == '__main__':
    step.start()
    time.sleep(1)
    step.shutdown()



import pytest

from onestep import DropMessage
from onestep.message import Message

from onestep.middleware import MemoryUniqueMiddleware


@pytest.fixture
def onestep():
    from onestep import step
    return step


def test_has_seen():
    mum = MemoryUniqueMiddleware()
    assert mum.has_seen("123") is False
    mum.mark_seen("123")
    assert mum.has_seen("123") is True


def test_unique_middleware():
    msg = Message(body="123")

    mum = MemoryUniqueMiddleware()
    mum.before_consume(None, msg)
    mum.after_consume(None, msg)

    try:
        mum.before_consume(None, msg)
    except Exception as e:
        assert isinstance(e, DropMessage)



import threading

from onestep import MemoryBroker, step


def cancel_consume(message):
    # when receive "over", cancel consume
    if message.body == "over":
        return True


memory_broker = MemoryBroker(cancel_consume=cancel_consume)


@step(from_broker=memory_broker)
def once_job(message):
    # do something
    print("receive", message.body)


if __name__ == '__main__':
    threading.Thread(target=step.start, kwargs={'block': True}).start()

    for i in range(2):
        memory_broker.publish(i)
    memory_broker.publish("over")



import time
from random import randint

from onestep.message import Message

from queue import Queue
import pytest

from onestep.broker.rabbitmq import RabbitMQBroker, RabbitMQConsumer


@pytest.fixture(scope="function")
def broker():
    broker = RabbitMQBroker(
        f"test-{randint(1, 1000)}",
        {
            "host": "localhost",
            "port": 5672,
            "username": "admin",
            "password": "admin",
            "ssl": False,
        }
    )
    yield broker
    broker.client.shutdown()


def test_mq_stream_broker_consume(broker):
    messages = [
        ('{"a": "b"}', {"a": "b"}),
        ('123', 123),
        ('{"body": 123}', 123)
    ]

    for message in messages:
        broker.publish(message[0])

    consumer = broker.consume()
    assert isinstance(consumer, RabbitMQConsumer)
    for message in messages:
        result = next(consumer)
        assert isinstance(result, Message)
        assert result.body == message[1]
        result.broker = broker
        result.confirm()


@pytest.fixture
def queue():
    return Queue()


def test_mq_consume_multi_messages(broker):
    broker.prefetch = 5  # mock prefetch
    for _ in range(10):
        broker.publish('{"a": "b"}')

    consumer = broker.consume()
    time.sleep(1)
    assert consumer.queue.qsize() == broker.prefetch


def test_requeue(broker):
    broker.publish('{"body": {"a1": "b2"}}')
    consumer = broker.consume()

    assert isinstance(consumer, RabbitMQConsumer)
    result = next(consumer)
    broker.requeue(result, is_source=True)

    consumer = broker.consume()
    result = next(consumer)
    assert result.body['a1'] == 'b2'

    broker.requeue(result, is_source=False)

    consumer = broker.consume()
    result = next(consumer)
    assert result.body['a1'] == 'b2'



from datetime import datetime, timedelta
from onestep import step, CronBroker, Cron

once_broker = CronBroker("* * * * * */3", once=True, body="仅仅执行一次")
# 使用自定义的开始时间（当前时间的前一天）
custom_start_time = datetime.now() - timedelta(days=1)
custom_broker = CronBroker(
    "* * * * * */3", body="使用自定义时间", start_time=custom_start_time
)

# 新增：DSL 与宏示例
dsl_every_30s = CronBroker(Cron.every(seconds=30), body="DSL 每 30 秒")
dsl_weekly = CronBroker(
    Cron.weekly(on=["mon", "fri"], at="10:30"), body="DSL 每周一/周五 10:30"
)
macro_workdays = CronBroker("@workdays", body="宏：工作日 00:00")
macro_every_1m = CronBroker("@every 1m", body="宏：每 1 分钟")


@step(
    from_broker=[once_broker, dsl_every_30s, dsl_weekly, macro_workdays, macro_every_1m]
)
def cron_task(message):
    print(message)
    return


if __name__ == "__main__":
    step.set_debugging()
    step.start(block=True)
    # step.shutdown()



import httpx

from onestep import step
from onestep.broker import MemoryBroker, RabbitMQBroker
from onestep.middleware import MemoryUniqueMiddleware

page_broker = MemoryBroker()
list_broker = RabbitMQBroker("spider.list", {"username": "admin", "password": "admin"})
detail_broker = RabbitMQBroker("spider.detail", {"username": "admin", "password": "admin"})


@step(to_broker=page_broker)
def build_task():
    """模拟创建10个任务"""
    yield from range(1, 11)


@step(from_broker=page_broker, to_broker=list_broker, workers=2)
async def crawl_list(message):
    async with httpx.AsyncClient() as client:
        resp = await client.get(f"https://httpbin.org/anything/{message.body}", timeout=None)
        url = resp.json().get("url")
        return url


@step(from_broker=list_broker, to_broker=detail_broker,
      workers=10, middlewares=[MemoryUniqueMiddleware()])
async def crawl_detail(message):
    async with httpx.AsyncClient() as client:
        resp = await client.get(message.body, timeout=None)
        return resp.json()


if __name__ == '__main__':
    build_task()
    step.set_debugging()
    step.start(block=True)



from typing import Generator

import pytest

from onestep import MemoryBroker, BaseMiddleware, DropMessage
from onestep.message import Message
from onestep.onestep import SyncOneStep
from onestep.worker import BaseWorker, ThreadWorker


@pytest.fixture
def step():
    return SyncOneStep(fn=lambda message: message)


@pytest.fixture
def broker():
    return MemoryBroker()


@pytest.fixture
def worker(step, broker):
    wt = BaseWorker(step, broker)
    yield wt


@pytest.fixture
def thread_worker(step, broker):
    wt = ThreadWorker(step, broker)
    yield wt


def test_receive_messages_shutdown(worker, broker):
    worker._shutdown = True
    assert isinstance(worker.receive_messages(), Generator)


def test_receive_messages(worker, broker):
    broker.publish({"a": "b"})
    for message in worker.receive_messages():
        assert isinstance(message, Message)
        assert message.body == {"a": "b"}
        break


def test_receive_messages_while_once(thread_worker, broker):
    broker.publish({"a": "b"})
    broker.once = True
    for message in thread_worker.receive_messages():
        assert message.body == {"a": "b"}
    assert thread_worker._shutdown is True


def test_receive_messages_while_not_once(thread_worker, broker):
    broker.publish({"a": "b"})
    broker.once = False
    for message in thread_worker.receive_messages():
        assert message.body == {"a": "b"}
        break
    assert thread_worker._shutdown is False


def test_receive_messages_consume(thread_worker, broker):
    broker.publish({"a": "b"})
    for message in thread_worker.receive_messages():
        assert message.body == {"a": "b"}
        break


def test_handle_message_consume_exception(step, thread_worker, broker):
    class test_middleware(BaseMiddleware):
        def before_consume(self, step, message, *args, **kwargs):
            message.drop = True
            raise DropMessage('drop message')

    step.middlewares = [test_middleware()]
    message = Message(body={"a": "b"})
    thread_worker.handle_message(message)
    assert message.drop is True


def test_handle_message(step, thread_worker, broker):
    class test_middleware(BaseMiddleware):
        def before_consume(self, step, message, *args, **kwargs):
            assert message.body == {"a": "b"}

        def after_consume(self, step, message, *args, **kwargs):
            assert message.body == {"a": "c"}

    step.middlewares = [test_middleware()]
    message = Message(body={"a": "b"})
    step.fn = lambda msg: msg.body.update({"a": "c"})
    thread_worker.handle_message(message)


def test__run_real_instance(step, thread_worker, broker):
    step.fn = lambda msg: msg.body.update({"a": "c"})
    message = Message(body={"a": "b"})
    thread_worker._run_real_instance(message)
    assert message.body == {"a": "c"}



from .base import (
    BaseBroker, BaseConsumer
)
from .memory import MemoryBroker, MemoryConsumer
from .webhook import WebHookBroker
from .rabbitmq import RabbitMQBroker, RabbitMQConsumer
from .mysql import MysqlBroker
from .sqs import SQSBroker, SQSConsumer, SNSBroker
from .redis import RedisStreamBroker, RedisPubSubBroker
from .cron import CronBroker


__all__ = [
    "BaseBroker", "BaseConsumer",
    "MemoryBroker", "MemoryConsumer",
    "WebHookBroker",
    "RabbitMQBroker", "RabbitMQConsumer",
    "MysqlBroker",
    "SQSBroker", "SQSConsumer", "SNSBroker",
    "RedisStreamBroker", "RedisPubSubBroker",
    "CronBroker"
]



from onestep import step, CronBroker
from onestep.worker import ThreadPoolWorker

cron_broker = CronBroker("* * * * * */3", body="hi cron")


@step(from_broker=cron_broker,
      workers=3,
      worker_class=ThreadPoolWorker)
def cron_task(message):
    print(message)
    return message


if __name__ == '__main__':
    step.set_debugging()
    step.start(block=True)
    # step.shutdown()



import pytest

from onestep import step
from onestep.broker import MemoryBroker


@pytest.fixture
def broker():
    return MemoryBroker()


def test_func(broker):

    @step(from_broker=broker, name="do_some_thing")
    def do_some_thing(message):
        ...
    assert do_some_thing.__step_params__ is not None

    assert do_some_thing.__step_params__["from_broker"] == broker
    assert do_some_thing.__step_params__["name"] == "do_some_thing"


from queue import Queue
from typing import Optional, Dict, Any, cast

try:
    from use_mysql import MysqlStore, Model, SQLModel
except ImportError:
    MysqlStore = None

from .base import BaseBroker, BaseConsumer
from ..message import Message


class MysqlBroker(BaseBroker):

    def __init__(
        self,
        params: Optional[Dict] = None,
        auto_create: Optional[bool] = True,
        *args,
        **kwargs
    ):
        if MysqlStore is None or Model is None:
            raise ImportError("未安装 use-mysql 依赖")

        super().__init__(*args, **kwargs)
        self.queue = Queue()
        params = params or {}
        auto_create = auto_create or False
        self.client = MysqlStore(generate_schemas=auto_create, **params)
        self.client.init()
        self._shutdown = False

    def send(self, message, *args, **kwargs):
        if not isinstance(message, Message):
            message = self.message_cls(body=message)
        return self.publish(message, *args, **kwargs)


    def publish(self, message: Any, **kwargs):
        body_raw = message.body if isinstance(message, Message) else message
        if not isinstance(body_raw, SQLModel):
            raise TypeError("MySQL Broker 仅支持 Model 类型的消息体")

        body = cast(SQLModel, body_raw)
        return self.client.create(body.__class__, **body.model_dump())

  
    def confirm(self, message: Message):
        return None

    def reject(self, message: Message):
        return None

    def requeue(self, message: Message, is_source=False):
        return None

    def shutdown(self):
        self._shutdown = True
        self.client.shutdown()


class MysqlConsumer(BaseConsumer):
    ...



import pytest
from urllib import request

from onestep import WebHookBroker


@pytest.fixture
def webhook_broker():
    broker = WebHookBroker("/push", port=8888)
    yield broker
    broker.shutdown()


def test_webhook_broker(webhook_broker):
    consumer = webhook_broker.consume()

    url = "http://localhost:8888/push"
    data = b'{"message": "Test webhook"}'
    req = request.Request(url, data=data, headers={'Content-Type': 'application/json'})
    response = request.urlopen(req)

    assert response.getcode() == 200

    assert consumer.queue.qsize() == 1
    assert consumer.queue.get() == data.decode("utf-8")



from onestep import step
from onestep.broker import MemoryBroker
from onestep.exception import RetryInQueue, DropMessage
from onestep.retry import AdvancedRetry

todo_broker = MemoryBroker()


@step(to_broker=todo_broker)
def build_todo_list():
    # mock data
    yield from [
        {
            "id": 1,
            "title": "todo1",
            "content": "todo1 content",
            "status": "todo",
        },
        {
            "id": 2,
            "title": "todo2",
            "content": "todo2 content",
            "status": "todo",
        },
        {
            "id": 3,
            "title": "todo3",
            "content": "todo3 content",
            "status": "todo",
        },
    ]


@step(from_broker=todo_broker, workers=1,
      retry=AdvancedRetry()
      )
def do_something(todo):
    print(todo)
    todo.body["status"] = "done"
    if todo.body["id"] == 2:
        todo.body["id"] = 21
        raise RetryInQueue("test requeue")
    elif todo.body["id"] == 3:
        raise DropMessage("test reject")
    else:
        return todo  # test confirm


if __name__ == '__main__':
    import time
    step.set_debugging()

    build_todo_list()
    step.start(block=False)
    time.sleep(5)
    step.shutdown()



from onestep import step
from onestep.broker import WebHookBroker

step.set_debugging()

# 对外提供一个webhook接口，接收外部的消息
webhook_broker = WebHookBroker(path="/push")


@step(from_broker=webhook_broker)
def waiting_messages(message):
    print("收到消息：", message)
    message.requeue()


if __name__ == '__main__':
    step.set_debugging()
    step.start(block=True)



import pytest
from onestep.broker import MemoryBroker


@pytest.fixture
def broker():
    return MemoryBroker(once=True)


def test_once(broker):
    consumer = broker.consume()
    broker.send("message 1")
    broker.send("message 2")
    broker.send("message 3")

    # Consume only one message
    consumed_messages = []
    for message in consumer:
        consumed_messages.append(message)
        if broker.once:
            break

    assert len(consumed_messages) == 1



import time

from onestep import step
from onestep.broker import RabbitMQBroker, MemoryBroker

step.set_debugging()

# 配置RabbitMQ队列及连接信息
rmq_broker = RabbitMQBroker(
    "test2",
    {
        "username": "admin",
        "password": "admin",
    }
)

# 模拟一个内存队列
todo_broker = MemoryBroker()


# todo_broker.send("1")


@step(from_broker=rmq_broker, to_broker=todo_broker, workers=3)
def build_todo_list(message):
    print("build_todo_list", message.body)
    # 返回的内容将发给RabbitMQ队列
    return message


if __name__ == '__main__':
    step.start()
    time.sleep(2)
    step.shutdown()



from onestep.retry import AllRetry, AnyRetry, RetryStatus


def test_all_retry():
    def retry1(message):
        return RetryStatus.CONTINUE

    def retry2(message):
        return RetryStatus.END_WITH_CALLBACK

    all_retry = AllRetry(retry1, retry2)

    assert all_retry("") == RetryStatus.END_WITH_CALLBACK
    all_retry = AllRetry(retry1, retry1)

    assert all_retry("") == RetryStatus.CONTINUE


# 测试 AnyRetry 类
def test_any_retry():
    def retry1(message):
        return RetryStatus.CONTINUE

    def retry2(message):
        return RetryStatus.END_WITH_CALLBACK

    any_retry = AnyRetry(retry1, retry2)
    assert any_retry("") == RetryStatus.CONTINUE

    any_retry = AnyRetry(retry1, retry1)
    assert any_retry("") == RetryStatus.CONTINUE

    any_retry = AnyRetry(retry2, retry2)
    assert any_retry("") == RetryStatus.END_WITH_CALLBACK



from onestep import step, MemoryBroker, BaseMiddleware

todo_broker = MemoryBroker()
todo_broker.send("hello world")


class MyMiddleware(BaseMiddleware):
    def __init__(self):
        self.x = 123

    def before_send(self, step, message, *args, **kwargs):
        print(self.x)

    def after_send(self, step, message, *args, **kwargs):
        pass

    def before_consume(self, step, message, *args, **kwargs):
        pass

    def after_consume(self, step, message, *args, **kwargs):
        pass


# 由于from_broker和to_broker都是todo_broker，
# 所以这个step会在todo_broker中循环执行
@step(from_broker=todo_broker,
      to_broker=todo_broker,
      workers=3,
      middlewares=[MyMiddleware()])
def do_something(todo):
    return todo


if __name__ == '__main__':
    # step.set_debugging()

    step.start(block=True)



from onestep.message import Extra, Message


def test_message_extra():
    extra = Extra(task_id='123', publish_time=1234567890.123, failure_count=1)
    extra.new_attr = 'new_attr'
    assert extra.task_id == '123'
    assert extra.publish_time == 1234567890.123
    assert extra.failure_count == 1
    assert extra.new_attr == 'new_attr'


def test_message_to_dict():
    extra = Extra(task_id='123', publish_time=1234567890.123, failure_count=1)
    message = Message(body={'key': 'value'}, extra=extra)
    expected_output = {'body': {'key': 'value'},
                       'extra': {'task_id': '123', 'publish_time': 1234567890.123, 'failure_count': 1}}
    assert message.to_dict() == expected_output


def test_message_replace():
    msg = Message(body={'key': 'value'}, extra={'task_id': '123', 'publish_time': 1234567890.123, 'failure_count': 0})
    msg = msg.replace(body={'new_key': 'new_value'})
    expected_dict = {'body': {'new_key': 'new_value'},
                     'extra': {'task_id': '123', 'publish_time': 1234567890.123, 'failure_count': 0}}
    assert msg.to_dict() == expected_dict


def test_message_set_exception():
    msg = Message(body={'key': 'value'}, extra={'task_id': '123', 'publish_time': 1234567890.123, 'failure_count': 0})
    try:
        raise ValueError('test')
    except Exception:
        msg.set_exception()
    # msg.set_exception(ValueError('test'))
    assert str(msg.exception) == 'test'
    assert msg.exception.exc_type is ValueError
    assert msg.failure_count == 1
    print(msg.to_json(True))



from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional, Tuple, Union, Type, List

from .exception import RetryInLocal, RetryInQueue
from .message import Message


class RetryStatus(Enum):
    CONTINUE = 1  # 继续（执行重试）
    END_WITH_CALLBACK = 2  # 结束（执行回调）
    END_IGNORE_CALLBACK = 3  # 结束（忽略回调）


class BaseRetry(ABC):

    @abstractmethod
    def __call__(self, *args, **kwargs) -> Optional[RetryStatus]:
        pass


class NeverRetry(BaseRetry):

    def __call__(self, message) -> Optional[RetryStatus]:
        return RetryStatus.END_WITH_CALLBACK


class AlwaysRetry(BaseRetry):

    def __call__(self, message) -> Optional[RetryStatus]:
        return RetryStatus.CONTINUE


class AllRetry(BaseRetry):
    def __init__(self, *retries: BaseRetry):
        self.retries: Tuple[BaseRetry, ...] = retries

    def __call__(self, message: Message) -> Optional[RetryStatus]:
        for retry in self.retries:
            status = retry(message)
            if status != RetryStatus.CONTINUE:
                return status
        return RetryStatus.CONTINUE


class AnyRetry(BaseRetry):
    def __init__(self, *retries: BaseRetry):
        self.retries: Tuple[BaseRetry, ...] = retries

    def __call__(self, message: Message) -> Optional[RetryStatus]:
        for retry in self.retries:
            status = retry(message)
            if status == RetryStatus.CONTINUE:
                return RetryStatus.CONTINUE
        return RetryStatus.END_WITH_CALLBACK


class TimesRetry(BaseRetry):

    def __init__(self, times: int = 3):
        self.times = times

    def __call__(self, message) -> Optional[RetryStatus]:
        if message.failure_count < self.times:
            return RetryStatus.CONTINUE
        return RetryStatus.END_WITH_CALLBACK


class RetryIfException(BaseRetry):

    def __init__(self, exceptions: Union[Type[Exception], Tuple[Type[Exception], ...]] = Exception):
        self.exceptions = (exceptions,) if not isinstance(exceptions, tuple) else exceptions

    def __call__(self, message) -> Optional[RetryStatus]:
        if isinstance(message.exception.exc_value, self.exceptions):
            return RetryStatus.CONTINUE
        return RetryStatus.END_WITH_CALLBACK


class AdvancedRetry(TimesRetry):
    """高级重试策略

    1. 本地重试：如果异常是 RetryInLocal 或 指定异常，且重试次数未达到上限，则本地重试，不回调
    2. 队列重试：如果异常是 RetryInQueue，且重试次数未达到上限，则入队重试，不回调
    3. 其他异常：如果异常不是 RetryInLocal 或 RetryInQueue 或 指定异常，则不重试，回调
    注：待重试的异常若继承自 RetryException，则可单独指定重试次数，否则默认为 3 次
    """

    def __init__(self, times: int = 3, exceptions: Union[None, Type[Exception], Tuple[Type[Exception], ...]] = None):
        super().__init__(times=times)
        exceptions_tuple = (exceptions,) if exceptions is not None and not isinstance(exceptions, tuple) else exceptions
        self.exceptions = (RetryInLocal, RetryInQueue) + (exceptions_tuple or ())

    def __call__(self, message: Message) -> Optional[RetryStatus]:
        if message.exception is not None and isinstance(message.exception.exc_value, self.exceptions):
            max_retry_times = getattr(message.exception.exc_value, "times", None) or self.times
            if message.failure_count < max_retry_times:
                if isinstance(message.exception.exc_value, RetryInQueue):
                    return RetryStatus.END_IGNORE_CALLBACK
                return RetryStatus.CONTINUE
            else:
                return RetryStatus.END_WITH_CALLBACK
        else:
            return RetryStatus.END_WITH_CALLBACK


# error callback
class BaseErrorCallback(ABC):

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass


class NackErrorCallBack(BaseErrorCallback):

    def __call__(self, message):
        message.reject()



# -*- coding: utf-8 -*-
import abc
import logging
import time
from queue import Queue, Empty
from typing import Any, Optional, List, Callable, Type

from onestep.middleware import BaseMiddleware
from onestep.exception import StopMiddleware
from onestep.message import Message

logger = logging.getLogger(__name__)


class BaseBroker:
    message_cls: Type[Message] = Message

    def __init__(self,
                 name: Optional[str] = None,
                 queue: Optional[Queue] = None,
                 middlewares: Optional[List[BaseMiddleware]] = None,
                 once: bool = False,
                 cancel_consume: Optional[Callable] = None,
                 send_retry_times: int = 3,
                 send_retry_delay: float = 1.0):
        """
        @param name: broker name
        @param queue: broker queue name
        @param middlewares: broker middlewares
        @param once: just run once
            if once is True, when broker receive a message, it will shutdown
        @param cancel_consume: cancel consume
            cancel_consume(message) -> bool
            if cancel_consume return True, broker will shutdown
        @param send_retry_times: message send retry times (default: 3)
        @param send_retry_delay: message send retry delay in seconds (default: 1.0)
        """
        self.queue = queue
        self.name = name or "broker"
        self.middlewares: List[BaseMiddleware] = []
        self.once = once
        self.cancel_consume = cancel_consume
        self.send_retry_times = send_retry_times
        self.send_retry_delay = send_retry_delay

        if middlewares:
            for middleware in middlewares:
                self.add_middleware(middleware)

    def add_middleware(self, middleware: BaseMiddleware):
        if not isinstance(middleware, BaseMiddleware):
            raise TypeError(f"middleware must be BaseMiddleware instance, not {type(middleware)}")
        self.middlewares.append(middleware)

    def send(self, message, *args, **kwargs):
        """对消息进行预处理，然后再发送"""
        step = kwargs.pop("step", None)
        if not isinstance(message, Message):
            message = self.message_cls(body=message)
        self.before_emit("send", message=message, step=step)

        # 消息发送重试机制
        result = None
        for attempt in range(self.send_retry_times):
            try:
                result = self.publish(message.to_json(), *args, **kwargs)
                break  # 发送成功，退出重试
            except Exception as e:
                if attempt < self.send_retry_times - 1:
                    # 使用指数退避策略
                    delay = self.send_retry_delay * (2 ** attempt)
                    logger.warning(
                        f"Message send failed (attempt {attempt + 1}/{self.send_retry_times}), "
                        f"retrying in {delay}s: {e}"
                    )
                    time.sleep(delay)
                else:
                    logger.error(
                        f"Message send failed after {self.send_retry_times} attempts: {e}"
                    )
                    raise

        self.after_emit("send", message=message, step=step)
        return result

    @abc.abstractmethod
    def publish(self, message: Any, *args, **kwargs):
        """
        将消息原样发布到 broker 中。如果当前Broker是Job的to_broker, 则必须实现此方法
        """
        raise NotImplementedError('Please implement in subclasses.')

    @abc.abstractmethod
    def consume(self, *args, **kwargs):
        """
        消费消息。如果当前Broker是Job的from_broker, 则必须实现此方法
        """
        raise NotImplementedError('Please implement in subclasses.')

    @abc.abstractmethod
    def confirm(self, message: Message):
        """确认消息"""
        raise NotImplementedError('Please implement in subclasses.')

    @abc.abstractmethod
    def reject(self, message: Message):
        """拒绝消息"""
        raise NotImplementedError('Please implement in subclasses.')

    @abc.abstractmethod
    def requeue(self, message: Message, is_source=False):
        """
        重发消息：先拒绝 再 重入
        is_source = False 重入使用消息的当前状态
        is_source = True 重入使用消息的初始状态
        """
        raise NotImplementedError('Please implement in subclasses.')

    def before_emit(self, signal, **kwargs):
        signal = "before_" + signal
        self._emit(signal, **kwargs)

    def after_emit(self, signal, **kwargs):
        signal = "after_" + signal
        self._emit(signal, **kwargs)

    def _emit(self, signal, **kwargs):
        for middleware in self.middlewares:
            if not hasattr(middleware, signal):
                continue
            try:
                params = dict(kwargs)
                params.setdefault("broker", self)
                params.setdefault("step", None)
                getattr(middleware, signal)(**params)
            except StopMiddleware:
                break

    def shutdown(self):
        """关闭Broker"""

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name}>"

    def __str__(self):
        return self.name


class BaseConsumer:

    def __init__(self, broker: BaseBroker, *args, **kwargs):
        self.queue = broker.queue
        self.message_cls: Type[Message] = broker.message_cls if broker.message_cls else Message
        self.timeout = kwargs.pop("timeout", 1000)

    def __next__(self):
        try:
            q = self.queue
            if q is None:
                return None
            timeout_ms = self.timeout
            if isinstance(timeout_ms, (int, float)) and timeout_ms > 0:
                data = q.get(timeout=timeout_ms / 1000.0)
            else:
                # 当超时为0、负数或非数字时，使用非阻塞获取以避免ValueError
                data = q.get_nowait()
            return self.message_cls.from_broker(broker_message=data)
        except (Empty, ValueError):
            return None

    def __iter__(self):
        return self



from pytest import fixture
from onestep.message import Message
from onestep import NeverRetry, AlwaysRetry, TimesRetry, RetryIfException, RetryInQueue
from onestep.retry import RetryStatus, AdvancedRetry


@fixture
def message():
    return Message({"a": 1})


def test_never_retry(message):
    retry_class = NeverRetry()
    assert RetryStatus.END_WITH_CALLBACK is retry_class(message)


def test_always_retry(message):
    retry_class = AlwaysRetry()
    assert RetryStatus.CONTINUE is retry_class(message)


def test_times_retry(message):
    retry_class = TimesRetry(3)
    assert RetryStatus.CONTINUE is retry_class(message)
    message.failure_count = 3
    assert RetryStatus.END_WITH_CALLBACK is retry_class(message)


def test_exception_retry2(message):
    retry_class = RetryIfException((ZeroDivisionError,))
    try:
        1 / 0
    except Exception:
        message.set_exception()
        assert RetryStatus.CONTINUE is retry_class(message)


def test_AdvancedRetry(message):
    retry_class = AdvancedRetry(times=3)
    try:
        1 / 0
    except Exception:
        message.set_exception()
    assert RetryStatus.END_WITH_CALLBACK is retry_class(message)

    retry_class = AdvancedRetry(times=3, exceptions=(ZeroDivisionError,))
    try:
        1 / 0
    except Exception:
        message.set_exception()
    assert RetryStatus.CONTINUE is retry_class(message)
    message.failure_count = 4
    assert RetryStatus.END_WITH_CALLBACK is retry_class(message)
    message.failure_count = 3
    assert RetryStatus.END_WITH_CALLBACK is retry_class(message)
    message.failure_count = 1
    assert RetryStatus.CONTINUE is retry_class(message)
    try:
        raise RetryInQueue()
    except Exception:
        message.set_exception()
    assert message.failure_count == 2
    assert RetryStatus.END_IGNORE_CALLBACK is retry_class(message)



import time
import json

import pytest

from onestep.broker.sqs.sqs import SQSBroker, SQSStore as RealSQSStore
from onestep.message import Message


class FakeMessage:
    def __init__(self, body, message_id=None):
        self.body = body if isinstance(body, str) else json.dumps(body)
        self.message_id = message_id or "msg-1"
        self.deleted = False

    def delete(self):
        self.deleted = True


class FakeSQSStore:
    def __init__(self, **kwargs):
        self._shutdown = False
        self._messages = []

    def declare_queue(self, queue_name, attributes=None, auto_create=True):
        return queue_name

    def send(self, queue_name, message, message_group_id=None, **kwargs):
        body = message if isinstance(message, (str, bytes)) else json.dumps(message)
        self._messages.append(FakeMessage(body))
        return message

    def start_consuming(self, queue_name, callback, prefetch=1, **kwargs):
        self._shutdown = False
        while not self._shutdown:
            count = 0
            while self._messages and count < prefetch:
                count += 1
                msg = self._messages.pop(0)
                callback(msg)
            time.sleep(0.05)

    def shutdown(self):
        self._shutdown = True


@pytest.fixture(autouse=True)
def patch_store(monkeypatch):
    monkeypatch.setattr("onestep.broker.sqs.sqs.SQSStore", FakeSQSStore)
    yield
    monkeypatch.setattr("onestep.broker.sqs.sqs.SQSStore", RealSQSStore)


def test_sqs_consume_and_confirm():
    broker = SQSBroker(queue_name="test-q.fifo", message_group_id="group", prefetch=2)
    broker.publish({"body": {"a": 2}})
    consumer = broker.consume()
    msg: Message = next(consumer)
    assert isinstance(msg, Message)
    assert msg.body == {"a": 2}
    broker.confirm(msg)
    broker.shutdown()


def test_sqs_requeue_source_and_current():
    broker = SQSBroker(queue_name="test-q.fifo", message_group_id="group", prefetch=1)
    broker.publish({"body": {"k": "v"}})
    consumer = broker.consume()
    msg = next(consumer)

    broker.requeue(msg, is_source=True)
    consumer2 = broker.consume()
    msg2 = next(consumer2)
    assert msg2.body == {"k": "v"}

    msg2.body = {"k": "v2"}
    broker.requeue(msg2, is_source=False)
    consumer3 = broker.consume()
    msg3 = next(consumer3)
    assert msg3.body == {"k": "v2"}
    broker.shutdown()


def test_sqs_single_consuming_thread():
    broker = SQSBroker(queue_name="test-q.fifo", message_group_id="group", prefetch=1)
    c1 = broker.consume()
    broker.consume()
    assert len(broker.threads) == 1
    broker.publish({"body": 123})
    time.sleep(0.1)
    assert c1.queue.qsize() >= 1
    broker.shutdown()


def test_sqs_prefetch():
    broker = SQSBroker(queue_name="test-q.fifo", message_group_id="group", prefetch=5)
    for i in range(10):
        broker.publish({"body": i})
    consumer = broker.consume()
    time.sleep(0.2)
    assert consumer.queue.qsize() >= 5
    broker.shutdown()


if __name__ == "__main__":
    test_sqs_consume_and_confirm()



from onestep import step, MemoryBroker
from onestep.middleware.config import RedisConsumeConfigMiddleware

from redis import Redis
from loguru import logger

rds = Redis(
    **{
        "host": "192.168.0.181",
        "port": 6379,
        "password": "password",
        "db": 0
    }
)


class MyConfigMiddleware(RedisConsumeConfigMiddleware):
    def process(self, step, message, *args, **kwargs):
        if message.body.get("site_name"):
            return self.get(message.body.get("site_name"))


# def
config_group = "onestep:config:fang.crawl"
config_middleware = MyConfigMiddleware(client=rds, config_group=config_group)
todo_broker = MemoryBroker()

# mock
cfg = '{ "domain": "nj", "region": { "areaId": null, "cityId": 659004882, "provinceId": 320000 }}'
rds.hset(config_group, "南京", cfg)
for i in range(3):
    todo_broker.send({"site_name": "南京", "say": f"hello world {i}"})


# 由于from_broker和to_broker都是todo_broker，
# 所以这个step会在todo_broker中循环执行
@step(
    from_broker=todo_broker,
    to_broker=todo_broker,
    workers=1,
    middlewares=[config_middleware])
def do_something(todo):
    logger.info(f"{todo=} {todo.config=}")


if __name__ == '__main__':
    # step.set_debugging()

    step.start(block=True)



# -*- coding: utf-8 -*-
"""
Cron DSL/Builder 与宏解析

提供人性化的 API 来生成 cron 表达式，并支持常用别名/宏：
- Cron.every(minutes=5) / Cron.every(seconds=10) / Cron.every(hours=2) / Cron.every(days=3) / Cron.every(months=1)
- Cron.daily(at="09:00")
- Cron.weekly(on="mon" or ["mon","fri"], at="10:30")
- Cron.monthly(on_day=1, at="00:00")
- Cron.yearly(on="01-01", at="00:00")
- 宏别名：@hourly/@daily/@weekly/@monthly/@yearly
- 扩展宏：@workdays（1-5）/@weekends（0,6）/@every <n><unit>（unit: s/m/h/d/mo）

注意：
- 标准 5 字段：minute hour day month day_of_week
- 当包含 seconds（如 `at` 使用 HH:MM:SS 或 `every(seconds=...)`）时输出 6 字段：second minute hour day month day_of_week
- `@every <n>d` 等按“日/月字段取模”的语义生成表达式，并非滚动间隔。
"""
from __future__ import annotations

from typing import List, Optional, Tuple, Union

DOW_MAP = {
    "sun": 0, "mon": 1, "tue": 2, "wed": 3, "thu": 4, "fri": 5, "sat": 6,
}

ALIAS_MAP = {
    "@hourly": "0 * * * *",
    "@daily": "0 0 * * *",
    "@weekly": "0 0 * * 0",
    "@monthly": "0 0 1 * *",
    "@yearly": "0 0 1 1 *",
}


def _parse_at(at: Optional[str]) -> Tuple[Optional[int], int, int]:
    """解析 at 参数，支持 HH:MM 或 HH:MM:SS。
    返回 (second, minute, hour)。如果不传 at，默认 minute=0, hour=0。
    """
    if not at:
        return None, 0, 0
    parts = at.strip().split(":")
    if len(parts) == 2:
        h, m = parts
        return None, int(m), int(h)
    elif len(parts) == 3:
        h, m, s = parts
        return int(s), int(m), int(h)
    raise ValueError(f"Invalid time format for at='{at}', expected HH:MM or HH:MM:SS")


def _to_dow_expr(on: Union[str, int, List[Union[str, int]]]) -> str:
    """将星期输入转换为 day_of_week 字段表达式。
    支持：字符串缩写（mon...sun）、数字 0-6、列表。
    """
    def _to_num(v: Union[str, int]) -> int:
        if isinstance(v, int):
            if v < 0 or v > 6:
                raise ValueError("day_of_week must be in 0-6")
            return v
        v = str(v).lower()
        if v.isdigit():
            num = int(v)
            if num < 0 or num > 6:
                raise ValueError("day_of_week must be in 0-6")
            return num
        if v not in DOW_MAP:
            raise ValueError(f"Unknown day_of_week '{v}'")
        return DOW_MAP[v]

    if isinstance(on, (list, tuple, set)):
        vals = sorted({_to_num(v) for v in on})
        return ",".join(str(x) for x in vals)
    else:
        return str(_to_num(on))


def _join_fields(second: Optional[int], minute: Union[int, str], hour: Union[int, str], day: Union[int, str], month: Union[int, str], dow: Union[int, str]) -> str:
    """组合 5/6 字段。若 second 为 None，输出 5 字段，否则输出 6 字段（秒在最后）。"""
    if second is None:
        return f"{minute} {hour} {day} {month} {dow}"
    return f"{minute} {hour} {day} {month} {dow} {second}"


class Cron:
    @staticmethod
    def every(*, seconds: Optional[int] = None, minutes: Optional[int] = None, hours: Optional[int] = None, days: Optional[int] = None, months: Optional[int] = None) -> str:
        """每隔指定单位执行。
        只能指定一个非 None 的单位。
        - seconds: 6 字段：*/s * * * * *
        - minutes: 5 字段：*/m * * * *
        - hours:   5 字段：0 */h * * *
        - days:    5 字段：0 0 */d * *
        - months:  5 字段：0 0 1 */mo *
        """
        units = [("seconds", seconds), ("minutes", minutes), ("hours", hours), ("days", days), ("months", months)]
        chosen = [(k, v) for k, v in units if v is not None]
        if len(chosen) != 1:
            raise ValueError("Cron.every() must specify exactly one unit")
        k, v = chosen[0]
        if not isinstance(v, int) or v <= 0:
            raise ValueError("Interval must be a positive integer")
        if k == "seconds":
            return f"* * * * * */{v}"
        if k == "minutes":
            return "*/%d * * * *" % v
        if k == "hours":
            return "0 */%d * * *" % v
        if k == "days":
            return "0 0 */%d * *" % v
        if k == "months":
            return "0 0 1 */%d *" % v
        raise RuntimeError("unreachable")

    @staticmethod
    def daily(*, at: Optional[str] = None) -> str:
        s, m, h = _parse_at(at)
        day, month, dow = "*", "*", "*"
        return _join_fields(s, m, h, day, month, dow)

    @staticmethod
    def weekly(*, on: Union[str, int, List[Union[str, int]]] = "mon", at: Optional[str] = None) -> str:
        s, m, h = _parse_at(at)
        dow = _to_dow_expr(on)
        day, month = "*", "*"
        return _join_fields(s, m, h, day, month, dow)

    @staticmethod
    def monthly(*, on_day: Union[int, List[int]] = 1, at: Optional[str] = None) -> str:
        s, m, h = _parse_at(at)
        month, dow = "*", "*"
        def _to_day(v: Union[int, str]) -> int:
            iv = int(v)
            if iv < 1 or iv > 31:
                raise ValueError("day_of_month must be in 1-31")
            return iv
        if isinstance(on_day, (list, tuple, set)):
            day = ",".join(str(_to_day(v)) for v in sorted({int(x) for x in on_day}))
        else:
            day = str(_to_day(on_day))
        return _join_fields(s, m, h, day, month, dow)

    @staticmethod
    def yearly(*, on: str = "01-01", at: Optional[str] = None) -> str:
        s, m, h = _parse_at(at)
        dow = "*"
        try:
            mon_str, day_str = on.split("-")
            month = str(int(mon_str))
            day = str(int(day_str))
        except Exception:
            raise ValueError("yearly(on=...) expects 'MM-DD', e.g., '01-01'")
        return _join_fields(s, m, h, day, month, dow)

    @staticmethod
    def alias(name: str) -> str:
        """返回内置宏别名对应的表达式（如果已知）。未知别名原样返回，以便 croniter 自行处理。"""
        return ALIAS_MAP.get(name.strip().lower(), name)


def resolve_cron(cron_like: str) -> str:
    """将类似 cron 的输入（字符串/宏）解析为标准 cron 表达式。
    - 内置关键字直接映射或保留（croniter 也支持）；
    - 扩展宏：@workdays/@weekends/@every <n><unit>（unit: s/m/h/d/mo）。
    """
    s = str(cron_like).strip()
    if not s.startswith("@"):
        return s
    lower = s.lower()
    # 内置别名
    if lower in ALIAS_MAP:
        return ALIAS_MAP[lower]
    # 扩展：工作日/周末（默认 00:00 执行）
    if lower == "@workdays":
        return "0 0 * * 1-5"
    if lower == "@weekends":
        return "0 0 * * 0,6"
    # 扩展：@every <n><unit>
    if lower.startswith("@every"):
        tail = lower.replace("@every", "", 1).strip()
        if not tail:
            raise ValueError("@every requires an interval, e.g., '@every 5m'")
        # 支持类似 "5m" / "10 h" / "2d" / "3mo" / "30s"
        tail = tail.replace(" ", "")
        # 处理 'mo' 与单位优先级
        if tail.endswith("mo"):
            num = int(tail[:-2])
            if num <= 0:
                raise ValueError("@every months must be > 0")
            return "0 0 1 */%d *" % num
        unit = tail[-1]
        try:
            num = int(tail[:-1])
        except Exception:
            raise ValueError("@every expects '<n><unit>', units: s/m/h/d/mo")
        if num <= 0:
            raise ValueError("@every interval must be > 0")
        if unit == "s":
            return "* * * * * */%d" % num
        if unit == "m":
            return "*/%d * * * *" % num
        if unit == "h":
            return "0 */%d * * *" % num
        if unit == "d":
            return "0 0 */%d * *" % num
        if unit == "w":
            # Cron 不支持“每 N 周”，仅建议直接使用 weekly(on=..., at=...)
            raise ValueError("@every <n>w is not supported; use weekly(on=..., at=...) instead")
        raise ValueError("Unknown unit for @every; use s/m/h/d/mo")
    # 未知宏直接返回，交由 croniter 处理（可能抛错）
    return s


from onestep.broker import MemoryBroker


def test_cancel_consume():
    def cancel_consume(message):
        if message.body == "cancel":
            return True
        return False

    broker = MemoryBroker(cancel_consume=cancel_consume)
    # Add messages to the queue
    broker.queue.put("message 1")
    broker.queue.put("cancel")
    broker.queue.put("message 2")

    consumed_messages = []
    for message in broker.consume():
        if broker.cancel_consume(message):
            break
        consumed_messages.append(message)

    assert len(consumed_messages) == 1
    assert consumed_messages[0].body == "message 1"



import pytest

from onestep import MemoryBroker
from onestep.onestep import BaseOneStep
from onestep.worker import ThreadWorker


@pytest.fixture
def broker():
    return MemoryBroker()


@pytest.fixture
def worker_thread(broker):
    onestep = BaseOneStep(fn=lambda message: message)
    wt = ThreadWorker(onestep, broker)
    wt.start()
    yield wt


def test_shutdown(worker_thread):
    worker_thread.shutdown()
    assert worker_thread._shutdown is True


def test_run_loop(worker_thread):
    worker_thread.shutdown()
    worker_thread.run()
    worker_thread.run()
    worker_thread.run()
    assert worker_thread._shutdown is True


def test_broker_once(worker_thread):
    worker_thread.broker.once = True
    worker_thread.broker.publish("test")
    worker_thread.run()
    assert worker_thread._shutdown is True



import json

import pytest

from onestep.broker import MemoryBroker
from onestep.middleware import BaseMiddleware
from onestep.onestep import SyncOneStep
from onestep.worker import ThreadWorker
from onestep.message import Message


class BrokerTestMiddleware(BaseMiddleware):
    def __init__(self):
        self.calls = []

    def before_send(self, step, message, *args, **kwargs):
        self.calls.append("before_send")

    def after_send(self, step, message, *args, **kwargs):
        self.calls.append("after_send")

    def before_consume(self, step, message, *args, **kwargs):
        self.calls.append("before_consume")
        message.broker_mark_consume = True

    def after_consume(self, step, message, *args, **kwargs):
        self.calls.append("after_consume")

    def before_confirm(self, step, message, *args, **kwargs):
        self.calls.append("before_confirm")

    def after_confirm(self, step, message, *args, **kwargs):
        self.calls.append("after_confirm")


@pytest.fixture
def broker():
    m = BrokerTestMiddleware()
    b = MemoryBroker(middlewares=[m])
    return b, m


def test_broker_send_triggers(broker):
    b, m = broker
    b.send({"a": 1})
    _ = b.queue.get()
    assert "before_send" in m.calls and "after_send" in m.calls


def test_broker_consume_triggers(broker):
    b, m = broker
    b.publish({"a": 2})

    step = SyncOneStep(fn=lambda msg: None)
    worker = ThreadWorker(step, b)
    msg = next(b.consume())
    worker.handle_message(msg)
    assert "before_consume" in m.calls and "after_consume" in m.calls
    assert getattr(msg, "broker_mark_consume", False) is True


def test_broker_confirm_triggers(broker):
    b, m = broker
    b.publish({"a": 3})
    msg = next(b.consume())
    msg.broker = b
    msg.confirm()
    assert "before_confirm" in m.calls and "after_confirm" in m.calls



import pytest
from onestep.message import Extra


@pytest.fixture
def extra_instance():
    # 创建一个 Extra 实例供测试使用
    return Extra()


def test_to_dict(extra_instance):
    # 测试 to_dict 方法是否返回正确的字典
    expected_dict = {
        'task_id': extra_instance.task_id,
        'publish_time': extra_instance.publish_time,
        'failure_count': extra_instance.failure_count,
    }
    assert extra_instance.to_dict() == expected_dict


def test_str(extra_instance):
    # 测试 __str__ 方法是否返回正确的字符串表示
    expected_str = str(extra_instance.to_dict())
    assert str(extra_instance) == expected_str


def test_custom_values():
    # 测试使用自定义值创建 Extra 实例是否正确
    task_id = 'custom_task_id'
    publish_time = 123456789.0
    failure_count = 2

    extra = Extra(task_id=task_id, publish_time=publish_time, failure_count=failure_count)
    assert extra.task_id == task_id
    assert extra.publish_time == publish_time
    assert extra.failure_count == failure_count


def test_default_values():
    # 测试使用默认值创建 Extra 实例是否正确
    extra = Extra()
    assert isinstance(extra.task_id, str)
    assert isinstance(extra.publish_time, float)
    assert extra.failure_count == 0



import time

from onestep import step
from onestep.broker import RabbitMQBroker

step.set_debugging()

# 配置RabbitMQ队列及连接信息
rmq_broker = RabbitMQBroker(
    "test2",
    {
        "username": "admin",
        "password": "admin",
    }
)


# for i in range(5):
#     rmq_broker.send(str(i))


@step(from_broker=rmq_broker, workers=3)
def do_some_thing(message):
    print("do_some_thing", message.body)
    time.sleep(3)


if __name__ == '__main__':
    step.start(block=True)



def test_rabbitmq_broker():
    from onestep.broker.rabbitmq import RabbitMQBroker
    from onestep.message import Message
    from loguru import logger
    
    rmq_broker = RabbitMQBroker(
        "test_one",
        {
            "username": "admin",
            "password": "admin",
        }
    )
    rmq_broker.send(message={"test": "test_one"})
    consumer = rmq_broker.consume()
    message: Message = next(consumer)
    logger.debug(f"message: {message}")
    message.broker = message.broker or rmq_broker
    message.confirm()
    logger.debug("message.confirm()")
    # logger.debug("del rmq_broker.client.connection")
    # del rmq_broker.client.connection
    # assert message.requeue(is_source=True) is None



"""
使用 CRON 表达式触发任务执行

支持人性化 DSL 与宏：
- DSL：Cron.every(minutes=5)、Cron.daily(at="09:00")、Cron.weekly(on=["mon","fri"], at="10:30") 等
- 宏：@hourly/@daily/@weekly/@monthly/@yearly 以及扩展 @workdays/@weekends/@every 5m/2h/3d/1mo
"""
import logging
import threading
from datetime import datetime
from typing import Any, Optional

from croniter import croniter  # type: ignore[import-untyped]
from ..cron import resolve_cron

from .memory import MemoryBroker, MemoryConsumer

logger = logging.getLogger(__name__)


class CronBroker(MemoryBroker):

    def __init__(self, cron: Any, name: Optional[str] = None, middlewares: Any = None, body: Any = None, start_time: Optional[datetime] = None, *args, **kwargs):
        super().__init__(name=name, middlewares=middlewares, *args, **kwargs)
        # 支持 DSL/宏/原始字符串：统一解析为标准表达式
        self.cron = resolve_cron(cron)
        self.start_time = start_time or datetime.now()
        self.itr = croniter(self.cron, self.start_time)
        self.next_fire_time = self.itr.get_next(datetime)
        self.body = body
        self._thread: Optional[threading.Timer] = None

    def _scheduler(self) -> None:
        if self.next_fire_time <= datetime.now():
            self.next_fire_time = self.itr.get_next(datetime)
            self.publish(self.body)

        self._thread = threading.Timer(interval=1, function=self._scheduler)
        self._thread.start()

    def consume(self, *args, **kwargs):
        self._scheduler()
        return CronConsumer(self, *args, **kwargs)

    def shutdown(self) -> None:
        if self._thread is not None:
            self._thread.cancel()


class CronConsumer(MemoryConsumer):
    ...
